//-----------------------------------------------------------------------------
// Copyright (c) 2007 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------

#if !defined(THREAD_H)
#define THREAD_H

#include <windows.h.>
#include <process.h>

//-----------------------------------------------------------------------------
// A Mutex allows threads mutually exclusive access to a resource.
//-----------------------------------------------------------------------------
class Mutex
{
public:
    Mutex()
    { InitializeCriticalSection(&m_mutex); }

    ~Mutex()
    { DeleteCriticalSection(&m_mutex); }

    void acquire()
    { EnterCriticalSection(&m_mutex); }

    void release()
    { LeaveCriticalSection(&m_mutex); }

private:
    CRITICAL_SECTION m_mutex;
};

//-----------------------------------------------------------------------------
// A Lock provides a safe way to acquire and release a Mutex. The Mutex 
// is acquired when the Lock it created. The Mutex is released when the
// Lock goes out of scope.
//-----------------------------------------------------------------------------
class Lock
{
public:
    Lock(Mutex &mutex) : m_mutex(mutex)
    { m_mutex.acquire(); }

    ~Lock()
    { m_mutex.release(); }

private:
    Mutex &m_mutex;
};

//-----------------------------------------------------------------------------
// Events are used to allow threads to communicate with each other. They are
// typically used to signal other threads when a certain 'event' has occurred.
//-----------------------------------------------------------------------------
class Event
{
public:
    Event()
    { m_hEvent = CreateEvent(0, FALSE, FALSE, 0); }

    ~Event()
    { CloseHandle(m_hEvent); }

    void notify()
    { SetEvent(m_hEvent); }

    void wait()
    { WaitForSingleObject(m_hEvent, INFINITE); }

private:
    HANDLE m_hEvent;
};

//-----------------------------------------------------------------------------
// The IRunnable interface represents the executable code that a thread will
// run when the thread is started via the Thread::start() method.
//-----------------------------------------------------------------------------
class IRunnable
{
public:
    virtual ~IRunnable() = 0 {}
    virtual void run() = 0;
};

//-----------------------------------------------------------------------------
// The thread is always created in a suspended state. The thread will not start
// until the client application explicitly starts the thread.
//
// There are 3 ways to supply the Thread class with the code to be run:
//  1. Subclass the Thread class and override the run() method. Create an
//     instance of the derived Thread class and call start().
//  2. Create an instance of the Thread class and call start() with the address
//     of a class that implements the Runnable interface.
//  3. Create an instance of the Thread class and call start() with the address
//     of a function that takes no arguments and returns void.
//-----------------------------------------------------------------------------
class Thread
{
public:
    enum Priority
    {
        PRIORITY_LOWEST,
        PRIORITY_BELOW_NORMAL,
        PRIORITY_NORMAL,
        PRIORITY_ABOVE_NORMAL,
        PRIORITY_HIGHEST
    };

    Thread();
    virtual ~Thread();

    bool create(unsigned int stackSize = 0);
    bool create(const char *pszThreadName, unsigned int stackSize = 0);

    unsigned int id() const;
    const char *name() const;
    
    void start();
    void start(void (*pfnThreadFunc)());
    void start(IRunnable *pRunnable);   
    
    void join();

    Priority getPriority() const;
    void setPriority(Priority priority);

protected:
    virtual void run();

private:
    static unsigned int __stdcall threadFunc(void *pArgs);
    
    void setThreadName();

    HANDLE m_hThread;
    unsigned int m_threadId;
    const char *m_pszThreadName;
    IRunnable *m_pRunnable;
    void (*m_pfnThreadFunc)();
};

#endif